package packageAPI;

//import javacardx.crypto.*;
import javacard.security.*;
import javacard.framework.APDU;
import javacard.framework.Applet;
import javacard.framework.ISO7816;
import javacard.framework.ISOException;

public class API extends Applet {
	// Attributs
	private APIKey[] keys; 				// Tableau de clefs
	private boolean hasBeenInit; 		// Boolean si API initialisee
	private short 	clock; 				// Simulation d'horloge
	private short 	SHORT_MAX = 32767;	// valeur max de short avant de boucler à -x

	// Constante API
	public static final byte CLA_MONAPPLET = (byte) 0xAE;
	public static final byte LVL_MAX = (byte) 127;

	// Constantes pour communiquer - Raccourcis de commandes
	public static final byte INS_ADMIN_INIT = (byte) 0xA0;
	public static final byte INS_CREATE_KEY = (byte) 0xC0;
	public static final byte INS_DELETE_KEY = (byte) 0xDE;
	public static final byte INS_DECRYPT = (byte) 0xDC;
	public static final byte INS_ENCRYPT = (byte) 0xEC;
	public static final byte INS_READ_KEY = (byte) 0xF0;
	public static final byte INS_GET_CLOCK = (byte) 0xCA;
	public static final byte INS_TICK_CLOCK = (byte) 0xCB;
	public static final byte INS_RESET_CLOCK = (byte) 0xCC;
	public static final byte INS_ADMIN_CREATE = (byte) 0xA1;

	// Constructeur
	private API() {
		keys = null;
		hasBeenInit = false;
		clock = 0;
	}

	public static void install(byte bArray[], short bOffset, byte bLength)
			throws ISOException {
		new API().register();
	}

	// Fonctionnement de l'API
	public void process(APDU apdu) throws ISOException {
		// Recupere l'APDU sous forme d'un tableau d'octets
		byte[] buffer = apdu.getBuffer();

		// On verifie que l'APDU est en cours de sélection
		if (this.selectingApplet())
			return;

		byte CLA = (byte) (buffer[ISO7816.OFFSET_CLA] & 0xFF);
		byte INS = (byte) (buffer[ISO7816.OFFSET_INS] & 0xFF);

		// ISOException.throwIt permet de retourner le status word passé en
		// paramètre
		if (CLA != CLA_MONAPPLET) {
			ISOException.throwIt(ISO7816.SW_CLA_NOT_SUPPORTED);
		}

		switch (INS) {

		case INS_ENCRYPT:
			encrypt(apdu);
			break;

		case INS_DECRYPT:
			decrypt(apdu);
			break;

		case INS_CREATE_KEY:
			generateKey(apdu);
			break;

		case INS_ADMIN_INIT:
			adminInit(apdu);
			break;

		case INS_DELETE_KEY:
			deleteKey(apdu);
			break;

		case INS_READ_KEY:
			readKey(apdu);
			break;

		case INS_GET_CLOCK:
			returnClock(apdu);
			break;

		case INS_TICK_CLOCK:
			tickTheClock();
			break;

		case INS_RESET_CLOCK:
			resetClock();
			break;

		case INS_ADMIN_CREATE:
			create(apdu);
			break;

		default:
			ISOException.throwIt(ISO7816.SW_INS_NOT_SUPPORTED);
		}
	}

	// Methode pour l'initialisation administrateur
	private void adminInit(APDU apdu) throws ISOException, CryptoException {
		// On teste si l'API a deja ete initialisee
		if (hasBeenInit) {
			ISOException.throwIt(ISO7816.SW_FUNC_NOT_SUPPORTED);
			return;
		} else {
			// On previent l'APi qu'il y a des choses a recevoir dans l'APDU
			// NB : C'est super important pour la carte reelle !!!!
			apdu.setIncomingAndReceive();

			// On initialise le tableau de clefs
			keys = new APIKey[50];
			for (short i = 0; i < keys.length; i++) {
				keys[i] = new APIKey();
				// NB : Clefs 3DES = 192 bits (0xC0), donc 24 octets.
			}

			// On recupere une nouvelle fois l'APDU
			byte[] buffer = apdu.getBuffer();
			// On recupere le nombre de clefs
			byte n = (byte) (buffer[ISO7816.OFFSET_CDATA] & 0xFF);

			// Le premier octet du buffer donnera le nombre de clefs admin a
			// initialiser
			// Chaque clef est accompagnee de son level sur le schema 1 oct
			// (level) + 24 octs (clef)

			if ((byte) 0x00 < n && n < (byte) 0x05) {
				// On va parcourir les clefs de l'APDU
				for (short i = 0x00; i < n; i++) {
					// On initialise les clefs de l'API (Commencant forcement a
					// 0)
					keys[i].init3DES(
					// On copie a partir de buffer...
							buffer,
							// On definit l'indice a partir duquel copier...
							// +1 pour sauter le nombre de clef et arriver au
							// level
							// +1 pour sauter le level et arriver au premier
							// byte de la clef
							// +25 (0x19) pour sauter une clef entiere
							(short) (ISO7816.OFFSET_CDATA + 0x02 + i * 0x19),
							// On donne le level
							buffer[(short) (ISO7816.OFFSET_CDATA + 0x01 + i * 0x19)],
							// On donne le type (type ADMIN = level ADMIN)
							buffer[(short) (ISO7816.OFFSET_CDATA + 0x01 + i * 0x19)],
							// On donne la date limite de péremption de la clé admin
							getMaxClock(buffer[(short) (ISO7816.OFFSET_CDATA + 0x01 + i * 0x19)]));
				}
				// On met en place la reponse... (les handles a retourner)
				// On recupere la taille maximale de la reponse
				short le = apdu.setOutgoing();
				if (le < n) {
					hasBeenInit = false;
					ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);
				} else {
					// On remplit le buffer avec les handles qui sont les
					// premiers indices...
					for (short i = 0x00; i < n; i++) {
						buffer[i] = (byte) i;
					}
					// On envoie les indices
					apdu.setOutgoingLength(n);
					apdu.sendBytes((short) 0, n);
				}
			} else {
				hasBeenInit = false;
				ISOException.throwIt(ISO7816.SW_WRONG_P1P2);
			}
			hasBeenInit = true;
			return;
		}
	}

	// Methode pour recuperer le contenu d'un handle et le renvoyer (TEST only)
	private void readKey(APDU apdu) throws ISOException {
		// On teste si l'API a deja ete initialisee
		if (!hasBeenInit) {
			ISOException.throwIt(ISO7816.SW_FUNC_NOT_SUPPORTED);
			return;
		}
		// On previent l'APi qu'il y a des choses a recevoir dans l'APDU
		// NB : C'est super important pour la carte reelle !!!!
		apdu.setIncomingAndReceive();

		// On recupere les donnes de l'APDU
		byte[] buffer = apdu.getBuffer();
		// On recupere la taille attendue de la reponse
		short le = apdu.setOutgoing();
		// On recupere le numero du handle qui nous interesse
		byte n = buffer[ISO7816.OFFSET_CDATA];
		if (!keys[n].isInitialized()) {
			ISOException.throwIt(ISO7816.SW_FILE_NOT_FOUND);
			return;
		}
		// On copie l'information desiree dans le buffer
		if (le >= keys[n].getLength()) {
			// On redefinit la taille du buffer pour qu'il contienne les infos
			// d'une clef
			buffer = new byte[0x19];
			// On met le level
			buffer[0] = keys[n].getLevel();
			// On met la valeur de la clef
			for (byte i = 0; i < keys[n].getLength(); i++) {
				buffer[(short) (i + 1)] = keys[n].getKey(i);
			}
			// On envoie l'APDU nouvellement fait
			apdu.setOutgoingLength((byte) 0x19);
			apdu.sendBytesLong(buffer, (byte) 0, (byte) 0x19);
		} else {
			ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);
		}
	}

	// Methode pour supprimer le contenu d'un handle (TEST only)
	private void deleteKey(APDU apdu) throws ISOException {
		// On teste si l'API a deja ete initialisee
		if (!hasBeenInit) {
			ISOException.throwIt(ISO7816.SW_FUNC_NOT_SUPPORTED);
			return;
		}
		// On previent l'APi qu'il y a des choses a recevoir dans l'APDU
		// NB : C'est super important pour la carte reelle !!!!
		apdu.setIncomingAndReceive();
		// On recupere les donnes de l'APDU
		byte[] buffer = apdu.getBuffer();
		// On recupere la taille attendue de la reponse
		short le = apdu.setOutgoing();
		// On recupere le numero du handle qui nous interesse
		byte n = buffer[ISO7816.OFFSET_CDATA];
		if (!keys[n].isInitialized()) {
			ISOException.throwIt(ISO7816.SW_FILE_NOT_FOUND);
			return;
		}
		// On supprime la clef
		keys[n].clearKey();

		// On copie l'information desiree dans le buffer
		buffer = new byte[1];
		buffer[0] = (byte) 0;
		if (le >= buffer.length) {
			// On envoie l'APDU nouvellement fait
			apdu.setOutgoingLength((short) buffer.length);
			apdu.sendBytesLong(buffer, (byte) 0, (short) buffer.length);
		} else {
			ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);
		}
	}

	// Methode pour trouver un emplacement libre dans l'API
	private byte findFreeKey() throws ISOException {
		for (byte i = 0; i < keys.length; i++) {
			// On trouve de la place
			if (!keys[i].isInitialized()) {
				return i;
			}
		}
		// Pas de place, on renvoie SW_FILE_FULL
		ISOException.throwIt(ISO7816.SW_FILE_FULL);
		return -1;
	}

	// Methode pour generer une clef
	private void generateKey(APDU apdu) {
		// On teste si l'API a deja ete initialisee
		if (!hasBeenInit) {
			ISOException.throwIt(ISO7816.SW_FUNC_NOT_SUPPORTED);
			return;
		}
		// On previent l'APi qu'il y a des choses a recevoir dans l'APDU
		// NB : C'est super important pour la carte reelle !!!!
		apdu.setIncomingAndReceive();
		// On recupere le contenu de l'APDU
		byte[] buffer = apdu.getBuffer();

		// On teste le niveau
		if (0 >= buffer[ISO7816.OFFSET_CDATA]
				|| buffer[ISO7816.OFFSET_CDATA] >= LVL_MAX) {
			ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
			return;
		}

		// On cherche une place de libre dans l'API (la methode gere
		// l'exception)
		byte i = findFreeKey();

		// Generation de la clef
		byte[] alea = new byte[24];
		RandomData random = RandomData
				.getInstance(RandomData.ALG_PSEUDO_RANDOM); // On cree le
															// generateur
		random.setSeed(buffer, (short) 0, (short) 24); 		// On l'intialise
		random.generateData(alea, (short) 0, (short) 24); 	// On genere l'alea
		// on genere la clef avec l'alea 
		keys[i].init3DES(alea, 								
				(short) 0, buffer[ISO7816.OFFSET_CDATA], 	// le level
				buffer[ISO7816.OFFSET_CDATA + 1], 			// le type
				getMaxClock(buffer[ISO7816.OFFSET_CDATA]) );// date de péremption

		// Construction et envoi de la réponse.
		buffer[0] = i;
		apdu.setOutgoingAndSend((short) 0, (short) 1);
	}

	// Methode pour chiffrer une des informations
	private void encrypt(APDU apdu) {
		// On teste si l'API a deja ete initialisee
		if (!this.hasBeenInit) {
			ISOException.throwIt(ISO7816.SW_FUNC_NOT_SUPPORTED);
			return;
		}
		// On previent l'APi qu'il y a des choses a recevoir dans l'APDU
		// NB : C'est super important pour la carte reelle !!!!
		apdu.setIncomingAndReceive();
		// On recupere le contenu de l'APDU
		byte[] buffer = apdu.getBuffer();
		short ind_buf = (short) ISO7816.OFFSET_CDATA;
		// On recupere la taille attendue de la reponse
		short le = apdu.setOutgoing();

		// On recupere le handle de la clef qui va servir au chiffrement
		byte key_handle = buffer[ind_buf];
		ind_buf++;
		// On va pouvoir faire les tests sur la clef de chiffrement
		if (!keys[key_handle].isInitialized()
				|| keys[key_handle].getLevel() == LVL_MAX) {
			ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
			return;
		}

		// La clef est bonne on peut essayer de chiffrer...
		// On recupere le type et le nombre des donnees
		byte nbr_handle = buffer[ind_buf];
		ind_buf++;
		byte nbr_mess = buffer[ind_buf];
		ind_buf++;

		// Maintenant on peut traiter les donnees...
		byte[] donnees = new byte[(short) (1 + (nbr_handle + nbr_mess) * 28)];
		short ind_don = (short) 0;
		donnees[ind_don] = (byte) (nbr_handle + nbr_mess);
		ind_don++;
		// On recupere les infos en commencant par les handles, s'il y en a
		if (nbr_handle != (byte) 0) {
			for (short i = 0; i < nbr_handle; i++) {
				// On test si le handle est bon...
				if (!keys[buffer[ind_buf]].isInitialized()) {
					ISOException.throwIt(ISO7816.SW_FILE_NOT_FOUND);
					return;
				}
				// On recupere le level d'abord... (en faisant aussi le test au
				// passage)
				if (keys[key_handle].getLevel() <= keys[buffer[ind_buf]]
						.getLevel()) {
					ISOException.throwIt(ISO7816.SW_FILE_INVALID);
					return;
				}
				donnees[ind_don] = keys[buffer[ind_buf]].getLevel();
				ind_don++;
				// On récupère le type
				donnees[ind_don] = keys[buffer[ind_buf]].getType();
				ind_don++;
				// on récupère la date de péremption
				// -> big endian
				donnees[ind_don] = (byte) ((keys[buffer[ind_buf]].getMaxClock() >> 8) & 0xFF);
				ind_don++;
				// -> low endian
				donnees[ind_don] = (byte) (keys[buffer[ind_buf]].getMaxClock() & 0xFF);
				ind_don++;
				// On recupere la valeur de la clef...
				for (short j = 0; j < keys[buffer[ind_buf]].getLength(); j++) {
					donnees[ind_don] = keys[buffer[ind_buf]].getKey()[j];
					ind_don++;
				}
				// Une fois la clef copiee, on passe a la suivante...
				ind_buf++;
			}
		}
		// On recupere les message s'il y en a
		if (nbr_mess != (byte) 0) {
			for (short i = 0; i < nbr_mess; i++) {
				// On fixe le level d'abord...
				donnees[ind_don] = 0;
				ind_don++;
				// On fixe le type
				donnees[ind_don] = 0;
				ind_don++;
				// On fixe la date de péremption
				donnees[ind_don] = 0;
				ind_don++;
				donnees[ind_don] = 0;
				ind_don++;
				// On recupere la valeur du message...
				for (short j = 0; j < 24; j++) {
					donnees[ind_don] = buffer[ind_buf];
					ind_don++;
					ind_buf++;
				}
			}
		}

		// On peut maintenant passer au chiffrement
		byte[] chiffre = chiffrement(donnees, keys[key_handle].getKey());

		// On copie l'information desiree dans le buffer
		if (le >= chiffre.length) {
			// On envoie l'APDU nouvellement fait
			apdu.setOutgoingLength((short) chiffre.length);
			apdu.sendBytesLong(chiffre, (byte) 0, (short) chiffre.length);
		} else {
			ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);
		}
	}

	// Methode de "chiffrement"
	private byte[] chiffrement(byte[] donn, byte[] key) throws ISOException {
		byte[] resultat = new byte[(short) (donn.length + key.length + 1)];
		for (short i = 0; i < donn.length; i++) {
			resultat[i] = donn[i];
		}
		resultat[(short) donn.length] = (byte) ':';
		for (short i = 0; i < key.length; i++) {
			resultat[(short) (1 + donn.length + i)] = key[i];
		}
		return resultat;
	}

	// Methode pour dechiffrer une des informations
	private void decrypt(APDU apdu) {
		// On teste si l'API a deja ete initialisee
		if (!this.hasBeenInit) {
			ISOException.throwIt(ISO7816.SW_FUNC_NOT_SUPPORTED);
			return;
		}
		// On previent l'APi qu'il y a des choses a recevoir dans l'APDU
		// NB : C'est super important pour la carte reelle !!!!
		apdu.setIncomingAndReceive();
		// On recupere le contenu de l'APDU
		byte[] buffer = apdu.getBuffer();
		short ind_buf = (short) ISO7816.OFFSET_CDATA;
		// On recupere la taille attendue de la reponse
		short le = apdu.setOutgoing();

		// On recupere le handle de la clef qui va servir au dechiffrement
		byte key_handle = buffer[ind_buf];
		ind_buf++;

		// On va pouvoir faire les tests sur la clef de chiffrement
		if (!keys[key_handle].isInitialized()
				|| keys[key_handle].getLevel() == LVL_MAX) {
			ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
			return;
		}

		// La clef semble bonne on peut essayer de dechiffrer...
		// On prepare le message chiffre
		byte[] donnees = new byte[(short) (buffer[ISO7816.OFFSET_LC] - 1)];
		for (short i = 0; i < donnees.length; i++) {
			donnees[i] = buffer[ind_buf];
			ind_buf++;
		}

		// On peut maintenant essayer le dechiffrement
		byte[] clair;
		try {
			clair = dechiffrement(donnees, keys[key_handle].getKey());
		} catch (Exception e) {
			ISOException.throwIt(ISO7816.SW_COMMAND_NOT_ALLOWED);
			return;
		}
		if (clair.length == 0) {
			ISOException.throwIt(ISO7816.SW_BYTES_REMAINING_00);
			return;
		}

		// On peut maintenant traiter les donnees...
		byte[] dechiffre = new byte[(short) clair.length]; // Tableau qui va
															// contenir l'info
															// finale a envoyer.
		short ind_dec = 0;

		short nbr_donn = clair[0]; // Nbre de donnees a recuperer (indique dans
									// le message en clair)
		short ind_clair = 1;

		short lvl_crnt; 		// Stocke le level du message courant traite
		byte type_crnt; 		// Stocke le type du message courant traite
		byte clock_bigendian; 	// Stocke la valeur big endian de la date de péremption
		byte clock_lowendian;	// Stocke la valeur low endian de la date de péremption
		byte[] msg_crnt = new byte[24]; // Stocke les 24 bits du message courant
										// traite

		for (short i = 0; i < nbr_donn; i++) {
			// On recupere le message
			// le level
			lvl_crnt = clair[ind_clair];
			ind_clair++;
			// le type
			type_crnt = clair[ind_clair];
			ind_clair++;
			// la date de péremption
			clock_bigendian = clair[ind_clair];
			ind_clair++;
			clock_lowendian = clair[ind_clair];
			ind_clair++;
			
			for (short j = 0; j < msg_crnt.length; j++) {
				msg_crnt[j] = clair[ind_clair];
				ind_clair++;
			}
			// On regarde si l'information est secrete ou non
			if (0 < lvl_crnt && lvl_crnt < keys[key_handle].getLevel()) {
				// Cas de l'info secrete
				// Il faut stocker l'information donc trouver une place libre
				// On cherche une place de libre dans l'API (la methode gere
				// l'exception)
				byte place = findFreeKey();

				// On initialise la clef avec la valeur recuperee
				keys[place].init3DES(msg_crnt, (short) 0, (byte) lvl_crnt,
						type_crnt, byteToShortClock(clock_bigendian, clock_lowendian) );

				// La reponse est uniquement constituee du handle
				dechiffre[ind_dec] = place;
				ind_dec++;
				// du level
				dechiffre[ind_dec] = (byte) lvl_crnt;
				ind_dec++;
				// et du type
				dechiffre[ind_dec] = type_crnt;
				ind_dec++;
			} else {
				if (lvl_crnt == 0) {
					// Cas de l'info publique
					/*
					 * On rajoute un -1 pour espacer les infos sur les clés des
					 * messages clairs NB : Ceci dépend de la façon dont sont
					 * cryptés les messages : d'abord les clés et ensuite les
					 * messages
					 */
					dechiffre[ind_dec] = -1;
					ind_dec++;
					for (short k = 0; k < msg_crnt.length; k++) {
						dechiffre[ind_dec] = msg_crnt[k];
						ind_dec++;
					}
				} else {
					ISOException.throwIt(ISO7816.SW_DATA_INVALID);
					return;
				}
			}
		}

		// On copie l'information desiree dans le buffer
		if (le >= dechiffre.length) {
			// On envoie l'APDU nouvellement fait
			apdu.setOutgoingLength((short) dechiffre.length);
			apdu.sendBytesLong(dechiffre, (byte) 0, (short) dechiffre.length);
		} else {
			ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);
		}
	}

	// Methode de "dechiffrement"
	private byte[] dechiffrement(byte[] donn, byte[] key) throws ISOException {
		byte[] resultat = new byte[(short) (donn.length - key.length - 1)];

		// Il faut verifier que la clef correspond au chiffre
		boolean good_key = true;
		for (short i = 0; i < key.length; i++) {
			if (donn[(short) (donn.length - 1 - i)] != key[(short) (key.length - 1 - i)])
				good_key = false;
		}
		if (good_key == false) {
			ISOException.throwIt(ISO7816.SW_SECURITY_STATUS_NOT_SATISFIED);
			return new byte[0];
		}

		// Le test est ok on peut "dechiffrer"
		for (short i = 0; i < resultat.length; i++) {
			resultat[i] = donn[i];
		}
		return resultat;
	}

	// Méthode ADMIN pour créer une ou plusieurs clé
	private void create(APDU apdu) {
		// On teste si l'API a deja ete initialisee
		if (!this.hasBeenInit) {
			ISOException.throwIt(ISO7816.SW_FUNC_NOT_SUPPORTED);
			return;
		}
		// On previent l'APi qu'il y a des choses a recevoir dans l'APDU
		// NB : C'est super important pour la carte reelle !!!!
		apdu.setIncomingAndReceive();
		// On recupere le contenu de l'APDU
		byte[] buffer = apdu.getBuffer();
		short ind_buf = (short) ISO7816.OFFSET_CDATA;
		// On recupere la taille attendue de la reponse
		short le = apdu.setOutgoing();

		/** Le buffer est de la forme (n, hn, .., h1, C) */
		// On recupere le nombre de handle servant au dechiffrement
		byte nbr_handle = buffer[ind_buf];
		byte ind_cles = (byte) (ind_buf + 1); // Indice de la première clé
		ind_buf += nbr_handle + 1; // Début du message chiffre

		// On vérifie que les n handles sont différents
		for (short j = ind_cles; j < (byte) (ind_cles + nbr_handle - 1); j++) {
			for (short i = (short) (j + 1); i < ind_cles + nbr_handle; i++) {
				if (buffer[j] == buffer[i]) {
					ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
					return;
				}
			}
		}

		// On prepare le message chiffre
		byte[] donnees = new byte[(short) (buffer[ISO7816.OFFSET_LC] - 1 - nbr_handle)];
		for (short i = 0; i < donnees.length; i++) {
			donnees[i] = buffer[ind_buf];
			ind_buf++;
		}

		// On peut maintenant essayer le dechiffrement
		byte[] clair = new byte[donnees.length];
		// On dechiffre nbr_handle fois le message chiffre
		for (short j = ind_cles; j < ind_cles + nbr_handle; j++) {
			byte key_handle = buffer[j];
			// On va pouvoir faire les tests sur la clef de chiffrement
			if (!keys[key_handle].isInitialized()
					|| keys[key_handle].getLevel() != LVL_MAX) {
				ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
				return;
			}
			// La clé est bonne, on déchiffre
			try {
				clair = dechiffrement(donnees, keys[key_handle].getKey());
				donnees = new byte[clair.length];
				for (short i = 0; i < donnees.length; i++) {
					donnees[i] = clair[i];
				}
			} catch (Exception e) {
				ISOException.throwIt(ISO7816.SW_COMMAND_NOT_ALLOWED);
				return;
			}
			if (clair.length == 0) {
				ISOException.throwIt(ISO7816.SW_BYTES_REMAINING_00);
				return;
			}
		}

		// On peut maintenant traiter les donnees...
		byte[] dechiffre = new byte[(short) clair.length]; // Tableau qui va
															// contenir l'info
															// finale a envoyer.
		short ind_dec = 0;

		short nbr_donn = clair[0]; // Nbre de donnees a recuperer (indique dans
									// le message en clair)
		short ind_clair = 1;

		short lvl_crnt; 		// Stocke le level du message courant traite
		byte type_crnt; 		// Stocke le type du message courant traite
		byte clock_bigendian; 	// Stock la valeur big endian de la date de péremption
		byte clock_lowendian;	// Stock la valeur low endian de la date de péremption
		byte[] msg_crnt = new byte[24]; // Stocke les 24 bits du message courant
										// traite

		for (short i = 0; i < nbr_donn; i++) {
			// On recupere le message
			// le level
			lvl_crnt = clair[ind_clair];
			ind_clair++;
			// le type
			type_crnt = clair[ind_clair];
			ind_clair++;
			// la date de péremption
			clock_bigendian = clair[ind_clair];
			ind_clair++;
			clock_lowendian = clair[ind_clair];
			ind_clair++;
			
			for (short j = 0; j < msg_crnt.length; j++) {
				msg_crnt[j] = clair[ind_clair];
				ind_clair++;
			}
			
			// On vérifie que la date de péremption n'est pas dépassée
			// ni qu'elle est supérieure à celle authorisée au max
			short date_per = byteToShortClock(clock_bigendian, clock_lowendian);
			if(date_per <= clock || date_per > getMaxClock((byte) lvl_crnt)) {
				ISOException.throwIt(ISO7816.SW_CONDITIONS_NOT_SATISFIED);
			}
			
			// On vérifie que l'information est secrete
			if (0 < lvl_crnt && lvl_crnt < LVL_MAX) {
				// Cas de l'info secrete
				// Il faut stocker l'information donc trouver une place libre
				// On cherche une place de libre dans l'API (la methode gere
				// l'exception)
				byte place = findFreeKey();

				// On initialise la clef avec la valeur recuperee
				keys[place].init3DES(msg_crnt, (short) 0, (byte) lvl_crnt,
						type_crnt, date_per);

				// La reponse est uniquement constituee du handle
				dechiffre[ind_dec] = place;
				ind_dec++;
				// du level
				dechiffre[ind_dec] = (byte) lvl_crnt;
				ind_dec++;
				// et du type
				dechiffre[ind_dec] = type_crnt;
				ind_dec++;
			}
		}

		// On copie l'information desiree dans le buffer
		if (le >= dechiffre.length) {
			// On envoie l'APDU nouvellement fait
			apdu.setOutgoingLength((short) dechiffre.length);
			apdu.sendBytesLong(dechiffre, (byte) 0, (short) dechiffre.length);
		} else {
			ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);
		}
	}

	// Méthode pour retourner l'horloge (big-endian)
	// Essayer sans les conversions (clock sur 2 bytes directement)
	private void returnClock(APDU apdu) {

		// On previent l'APi qu'il y a des choses a recevoir dans l'APDU
		// NB : C'est super important pour la carte reelle !!!!
		apdu.setIncomingAndReceive();
		apdu.getBuffer();
		short le = apdu.setOutgoing();
		byte[] buffer = new byte[2];
		// Encodage de l'horloge en big-endian
		buffer[0] = 1; // (byte) ((clock >> 8) & 0xFF); // Les 8 bits de poids fort
		// buffer[1] = 2; // (byte) (clock & 0xFF); // Les 8 bits de poids faible
		
		if (le >= buffer.length) {
			// On envoie l'APDU nouvellement fait
			apdu.setOutgoingLength((short) buffer.length);
			apdu.sendBytesLong(buffer, (byte) 0, (short) buffer.length);
		} else {
			ISOException.throwIt(ISO7816.SW_WRONG_LENGTH);
		} 
	}

	// incrémente l'horloge et supprime les clés périmés
	private void tickTheClock() {
		// augmente la valeur de l'horloge
		if( clock == SHORT_MAX )
			clock = 0;
		else
			clock++;

		// réinitialise les clés périmées
		for (byte i = 0; i < keys.length; i++) {
			if (keys[i].getMaxClock() < clock) {
				keys[i].clearKey();
			}
		}
	}

	// remet à zéro l'horloge
	private void resetClock() {
		clock = 0;
		// Attention, date de péremption des clés non modifiée
	}
	
	// Transforme la valeur de l'horloge de byte à short
	private short byteToShortClock( byte buffer_bigE, byte buffer_lowE ) {
		return (short) ((buffer_bigE << 8)+ buffer_lowE);
	}
	
	// Donne la date de péremption par rapport à l'horloge et au level
	private short getMaxClock( byte level ) {
		short date = (short) (clock + level);
		if(date < 0)
			return (short) (SHORT_MAX + clock + level);
		else
			return (short) (clock + level);
	}
}